from enum import Enum


class Middleware(Enum):
    BEFORE_QUERY = 'before_query'
    BEFORE_INSERT = 'before_insert'
    BEFORE_UPDATE = 'before_update'

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"Middleware.{self.name}"



from dataclasses import fields
from typing import List, Any, TypeVar, Type, Generic

from ._middlewares import get_middlewares
from ._models import models
from ._mysql_executor import select_one, select_many
from ._where import Where, Or
from .enums import Middleware

T = TypeVar("T")


class RawQuery:
    def __init__(self, table: str, database: str | None, ds_id: str):
        self._table = table
        self._database = database

        if self._table is None or self._table == "":
            raise ValueError("table is required")

        self._ds_id = ds_id
        self._where = Where()
        self._select_fields = []
        self._ignore_fields = []
        self._order_by = None
        self._limit = None
        self._offset = None
        self._distinct = False

        self._model: callable = models.get(
            ds_id=self._ds_id, database=self._database, table=self._table
        )
        self._model_fields = [f.name for f in fields(self._model)]

    def select(self, *select_fields, distinct=False) -> "RawQuery":
        model_fields = [f.name for f in fields(self._model)]
        for select_field in select_fields:
            if select_field not in model_fields:
                raise ValueError(f"invalid field [{select_field}]")

        self._select_fields = select_fields
        self._distinct = distinct
        return self

    def ignore(self, *ignore_fields) -> "RawQuery":
        model_fields = [f.name for f in fields(self._model)]
        for ignore_field in ignore_fields:
            if ignore_field not in model_fields:
                raise ValueError(f"invalid field [{ignore_field}]")

        self._ignore_fields = ignore_fields
        return self

    def check_field(self, field: str):
        if field not in self._model_fields:
            raise ValueError(f"invalid field [{field}]")

    def eq(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.lt(field, value)
        return self

    def le(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.le(field, value)
        return self

    def in_(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> "RawQuery":
        self.check_field(field)
        self._where.like(field, value)
        return self

    def or_(self, or_: Or) -> "RawQuery":
        self._where.or_(or_)
        return self

    def desc(self, *order_by) -> "RawQuery":
        self._order_by = [f"{field} desc" for field in order_by]
        return self

    def asc(self, *order_by) -> "RawQuery":
        self._order_by = [f"{field} asc" for field in order_by]
        return self

    def limit(self, limit: int) -> "RawQuery":
        self._limit = limit
        return self

    def offset(self, offset: int) -> "RawQuery":
        self._offset = offset
        return self

    def one(self) -> dict[str, Any] | None:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        self._apply_before_query_middlewares()

        sql, args = self._build_select()

        return select_one(self._ds_id, sql, args)

    def list(self) -> list[dict[str, Any]]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        self._apply_before_query_middlewares()

        sql, args = self._build_select()

        rows = select_many(self._ds_id, sql, args)
        if rows is None:
            return []
        return rows

    def page(self, page: int, page_size: int) -> tuple[List[dict[str, Any]], int]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        self._apply_before_query_middlewares()

        self._limit = page_size
        self._offset = (page - 1) * page_size

        sql, args = self._build_select()
        count = self._count()
        if count == 0:
            return [], count

        rows = select_many(self._ds_id, sql, args)
        if rows is None:
            return [], count
        return rows, count

    def _count(self) -> int:
        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f"SELECT COUNT(*) FROM {table}"
        args = ()
        tree = self._where.tree()
        if len(tree.conditions) > 0:
            exp, args = tree.parse()
            sql += " WHERE " + exp

        return select_one(self._ds_id, sql, args)["COUNT(*)"]

    def count(self) -> int:
        self._apply_before_query_middlewares()
        return self._count()

    def _build_select(self) -> tuple[str, tuple[Any, ...]]:
        select_fields = [
            field for field in self._select_fields if field not in self._ignore_fields
        ]

        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'SELECT {"DISTINCT " if self._distinct and self._select_fields else ""}{",".join(select_fields)} FROM {table}'
        args = ()
        tree = self._where.tree()
        if len(tree.conditions) > 0:
            exp, args = tree.parse()
            sql += " WHERE " + exp
        if self._order_by is not None:
            sql += f' ORDER BY {",".join(self._order_by)}'
        if self._limit is not None:
            sql += f" LIMIT {self._limit}"
        if self._offset is not None:
            sql += f" OFFSET {self._offset}"

        return sql, args

    def _apply_before_query_middlewares(self):
        middlewares = get_middlewares(Middleware.BEFORE_QUERY)
        if middlewares is None or len(middlewares) == 0:
            return
        for middleware in middlewares:
            if callable(middleware):
                middleware(self)


class Query(RawQuery, Generic[T]):
    def __init__(self, cls: Type[T], database: str | None, ds_id: str):
        if not hasattr(cls, "__table_name__"):
            raise ValueError("invalid model class")
        super().__init__(cls.__table_name__, database, ds_id)
        self._cls = cls

    def eq(self, field: str, value: Any) -> "Query":
        super().eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> "Query":
        super().ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> "Query":
        super().gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> "Query":
        super().ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> "Query":
        super().lt(field, value)
        return self

    def le(self, field: str, value: Any) -> "Query":
        super().le(field, value)
        return self

    def in_(self, field: str, value: Any) -> "Query":
        super().in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> "Query":
        super().l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> "Query":
        super().r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> "Query":
        super().like(field, value)
        return self

    def or_(self, or_: Or) -> "Query":
        super().or_(or_)
        return self

    def desc(self, *order_by) -> "Query":
        super().desc(*order_by)
        return self

    def asc(self, *order_by) -> "Query":
        super().asc(*order_by)
        return self

    def limit(self, limit: int) -> "Query":
        self._limit = limit
        return self

    def offset(self, offset: int) -> "Query":
        self._offset = offset
        return self

    def one(self) -> T | None:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        row = super().one()
        if row is None:
            return None
        return self._cls(**row)

    def list(self) -> List[T]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        rows = super().list()
        if len(rows) == 0:
            return rows
        return [self._cls(**row) for row in rows]

    def page(self, page: int, page_size: int) -> tuple[List, int]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        rows, count = super().page(page, page_size)
        if len(rows) == 0:
            return rows, count
        return [self._cls(**row) for row in rows], count


def query(cls: Type[T], database: str | None = None, ds_id: str = "default"):
    return Query(cls, database, ds_id)


def raw_query(table: str, database: str | None = None, ds_id: str = "default"):
    return RawQuery(table, database, ds_id)



from typing import Callable

from pydorm.enums import Middleware

before_query_middlewares: list[Callable[..., None]] = []
before_update_middlewares: list[Callable[..., None]] = []
before_insert_middlewares: list[Callable[[..., dict], None]] = []


def use_middleware(middleware_type: Middleware, middleware):
    if middleware_type == Middleware.BEFORE_QUERY:
        before_query_middlewares.append(middleware)
    elif middleware_type == Middleware.BEFORE_UPDATE:
        before_update_middlewares.append(middleware)
    elif middleware_type == Middleware.BEFORE_INSERT:
        before_insert_middlewares.append(middleware)
    else:
        raise ValueError(f'unsupported middleware type: {middleware_type}')


def get_middlewares(middleware_type: Middleware):
    if middleware_type == Middleware.BEFORE_QUERY:
        return before_query_middlewares
    elif middleware_type == Middleware.BEFORE_UPDATE:
        return before_update_middlewares
    elif middleware_type == Middleware.BEFORE_INSERT:
        return before_insert_middlewares
    else:
        raise ValueError(f'unsupported middleware type: {middleware_type}')



from dataclasses import field, make_dataclass

from loguru import logger

from ._datasources import get_datasource


class Models:
    def __init__(self):
        self._model_dict: dict[str, callable] = {}
        self._table_structure_dict: dict[str, list[TableField]] = {}

    def get_structure(
        self, ds_id: str, database: str | None, table: str
    ) -> list["TableField"]:
        ds = get_datasource(ds_id)
        if database is None:
            database = ds.get_default_database()

        self.get(ds_id, database, table)

        key = f"{ds_id}.{database}.{table}"
        return self._table_structure_dict.get(key, None)

    def get(self, ds_id: str, database: str | None, table: str) -> callable:
        ds = get_datasource(ds_id)
        if database is None:
            database = ds.get_default_database()

        key = f"{ds_id}.{database}.{table}"
        model = self._model_dict.get(key, None)
        if model is None:
            return self._load_structure(ds_id, database, table)
        else:
            return model

    def remove(self, ds_id: str, database: str | None, table: str) -> None:
        ds = get_datasource(ds_id)
        if database is None:
            database = ds.get_default_database()

        key = f"{ds_id}.{database}.{table}"
        if key in self._model_dict:
            del self._model_dict[key]
        if key in self._table_structure_dict:
            del self._table_structure_dict[key]

    def _load_structure(self, ds_id: str, database: str | None, table: str):
        ds = get_datasource(ds_id)
        conn = ds.create_connection()
        c = conn.cursor()
        key = f"{ds_id}.{database}.{table}"
        try:
            sql = f"show full columns from {database}.{table}"
            logger.debug(f"[{id(conn)}] {sql}")

            c.execute(sql)
            rows = c.fetchall()
            table_fields = []
            for row in rows:
                table_field = TableField(
                    field_=row["Field"],
                    type_=row["Type"],
                    null_=row["Null"],
                    key_=row["Key"],
                    default_=row["Default"],
                    extra=row["Extra"],
                    comment=row.get("Comment", ""),
                )
                table_fields.append(table_field)
            self._table_structure_dict[key] = table_fields
            fields = [
                (table_field.field_, any, field(default=None))
                for table_field in table_fields
            ]
            self._model_dict[key] = make_dataclass(key, fields=fields)
            return self._model_dict[key]
        finally:
            c.close()
            conn.commit()


class TableField:

    def __init__(self, field_, type_, null_, key_, default_, extra, comment):
        self.field_ = field_
        self.type_ = type_
        self.null_ = null_
        self.key_ = key_
        self.default_ = default_
        self.extra = extra
        self.comment = comment


models = Models()



from typing import Any

from loguru import logger
from pymysql.cursors import DictCursor

from ._tx import Tx
from ._tx import tx_context


def debug(conn, sql, args):
    logger.debug(f"[{id(conn)}] {sql}")
    if args is not None and len(args) > 0:
        logger.debug(f"### {args}")


def select_one(ds_id: str, sql: str, args: tuple[Any, ...]) -> dict[str, Any] | None:
    tx = _get_tx(ds_id)
    conn = tx.begin()
    cursor: DictCursor | None = None
    debug(conn, sql, args)
    try:
        cursor = conn.cursor()
        sql = sql.replace("?", "%s")
        result = cursor.execute(sql, args)
        if result is None:
            return None

        row = cursor.fetchone()
        if tx.is_auto_commit():
            conn.commit()
        return row
    except Exception as e:
        if tx.is_auto_commit():
            conn.rollback()
        raise e
    finally:
        if cursor is not None:
            cursor.close()
        if tx.is_auto_commit():
            conn.close()


def select_many(
    ds_id: str, sql: str, args: tuple[Any, ...]
) -> list[dict[str, Any]] | None:
    tx = _get_tx(ds_id)
    conn = tx.begin()
    cursor: DictCursor | None = None
    debug(conn, sql, args)
    try:
        cursor = conn.cursor()
        sql = sql.replace("?", "%s")
        result = cursor.execute(sql, args)
        if result is None:
            return None

        rows = cursor.fetchall()
        if rows is None:
            return []

        row_list = list(rows)
        if tx.is_auto_commit():
            conn.commit()
        return row_list
    except Exception as e:
        if tx.is_auto_commit():
            conn.rollback()
        raise e
    finally:
        if cursor is not None:
            cursor.close()
        if tx.is_auto_commit():
            conn.close()


def execute(ds_id: str, sql: str, args: tuple[Any, ...]) -> (int, int):
    tx = _get_tx(ds_id)
    conn = tx.begin()
    cursor: DictCursor | None = None
    debug(conn, sql, args)

    try:
        cursor = conn.cursor()
        sql = sql.replace("?", "%s")
        row_affected = cursor.execute(sql, args)
        last_row_id = cursor.lastrowid
        if tx.is_auto_commit():
            conn.commit()
        return row_affected, last_row_id
    except Exception as e:
        if tx.is_auto_commit():
            conn.rollback()
        raise e
    finally:
        if cursor is not None:
            cursor.close()
        if tx.is_auto_commit():
            conn.close()


def executemany(ds_id: str, sql: str, args: list[tuple[any, ...]]) -> int | None:
    tx = _get_tx(ds_id)
    conn = tx.begin()
    cursor: DictCursor | None = None
    debug(conn, sql, args)

    try:
        cursor = conn.cursor()
        sql = sql.replace("?", "%s")
        row_affected = cursor.executemany(sql, args)
        if tx.is_auto_commit():
            conn.commit()
        return row_affected
    except Exception as e:
        if tx.is_auto_commit():
            conn.rollback()
        raise e
    finally:
        if cursor is not None:
            cursor.close()
        if tx.is_auto_commit():
            conn.close()


def _get_tx(ds_id: str) -> Tx:
    tx = tx_context.get()
    if tx is None or not tx.is_valid() or ds_id != tx.ds_id():
        tx = Tx(ds_id, auto_commit=True)
    return tx



from pymysql.cursors import DictCursor

from ._datasources import get_datasource, default_datasource


def raw_select(sql: str, args: tuple[any, ...], ds_id: str | None = None):
    ds = default_datasource() if ds_id is None else get_datasource(ds_id)
    conn = ds.create_connection()
    try:
        cursor: DictCursor = conn.cursor()
        sql = sql.replace("?", "%s")
        result = cursor.execute(sql, args)
        if result is None:
            return None
        rows = cursor.fetchall()
        if rows is None:
            return []

        return list(rows)
    finally:
        conn.close()



from typing import Protocol, runtime_checkable


@runtime_checkable
class IDataSource(Protocol):
    def get_id(self) -> str: ...

    def get_default_database(self) -> str: ...

    def create_connection(self): ...



import pymysql
from loguru import logger
from pymysql.cursors import DictCursor


class MysqlDataSource:
    def __init__(self, ds_id, host, port, user, password, database):
        self._ds_id = ds_id
        self._dialect = "mysql"
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._default_database = database

    def get_id(self) -> str:
        return self._ds_id

    def get_default_database(self) -> str:
        return self._default_database

    def create_connection(self):
        conn = pymysql.connect(
            host=self._host,
            port=self._port,
            user=self._user,
            password=self._password.encode("utf-8"),
            database=self._default_database,
            cursorclass=DictCursor,
        )
        logger.info(f"create connection [{id(conn)}]")
        return conn



from dataclasses import fields
from typing import Any, TypeVar, Type

from ._middlewares import get_middlewares
from ._models import models
from ._mysql_executor import execute
from ._where import Where, Or
from .enums import Middleware

T = TypeVar("T")


class Update:
    def __init__(self, table: str, database: str | None, ds_id: str):
        self._table: str = table
        self._database = database
        self._ds_id = ds_id

        if self._table is None or self._table == "":
            raise ValueError("table is required")

        self._where = Where()
        self._update_fields: dict = {}

        self._model: callable = models.get(
            ds_id=self._ds_id, database=self._database, table=self._table
        )
        self._model_fields = [f.name for f in fields(self._model)]

    def check_field(self, field: str):
        if field not in self._model_fields:
            raise ValueError(f"invalid field [{field}]")

    def has_field(self, field: str):
        return field in self._model_fields

    def eq(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.lt(field, value)
        return self

    def le(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.le(field, value)
        return self

    def in_(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> "Update":
        self.check_field(field)
        self._where.like(field, value)
        return self

    def or_(self, or_: Or) -> "Update":
        self._where.or_(or_)
        return self

    def set(self, set_args: dict = None, **args) -> "Update":
        if set_args is None:
            set_args = {}

        set_args.update(args)

        valid_fields = {}
        for k, v in set_args.items():
            if self.has_field(k):
                valid_fields[k] = v

        if len(valid_fields) == 0:
            raise ValueError("valid fields is required")

        self._update_fields = valid_fields
        return self

    def update(self) -> int:
        if self._where.count() == 0:
            # danger operation
            raise ValueError("update all is not supported")

        if len(self._update_fields) == 0:
            raise ValueError("update fields is required")

        sql, args = self._build_update()

        affected, last_row_id = execute(self._ds_id, sql, args)
        return affected or 0

    def _build_update(self) -> tuple[str, tuple[Any, ...]]:
        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'UPDATE {table} SET {",".join([f"{k}=?" for k in self._update_fields.keys()])}'
        args = tuple(self._update_fields.values())

        exp, args2 = self._where.tree().parse()
        sql += " WHERE " + exp
        args += args2
        return sql, args

    def delete(self) -> int:
        if self._where.count() == 0:
            # danger operation
            raise ValueError("delete all is not supported")

        sql, args = self._build_delete()

        affected, last_row_id = execute(self._ds_id, sql, args)
        return affected or 0

    def _build_delete(self) -> tuple[str, tuple[Any, ...]]:
        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f"DELETE FROM {table}"

        exp, args = self._where.tree().parse()
        sql += " WHERE " + exp
        return sql, args

    def _apply_before_update_middlewares(self):
        middlewares = get_middlewares(Middleware.BEFORE_UPDATE)
        if middlewares is None or len(middlewares) == 0:
            return
        for middleware in middlewares:
            middleware(self)


def update(
    table_or_cls: str | Type[T], database: str | None = None, ds_id: str = "default"
) -> Update:
    if isinstance(table_or_cls, str):
        table = table_or_cls
    else:
        table = table_or_cls.__table_name__
    return Update(table, database, ds_id)



from dataclasses import fields
from typing import Any, TypeVar, Type

from ._middlewares import get_middlewares
from ._models import models
from ._mysql_executor import execute, executemany
from .enums import Middleware

T = TypeVar("T")


class Insert:
    def __init__(self, table: str, database: str | None, ds_id: str):
        self._table: str = table
        self._database = database
        self._ds_id = ds_id

        if self._table is None or self._table == "":
            raise ValueError("table is required")

        self._model: callable = models.get(
            ds_id=self._ds_id, database=self._database, table=self._table
        )

    def insert(self, data: any) -> (int, int):
        if data is None:
            raise ValueError("null data")

        data = self._process_data(data)
        self._apply_before_insert_middlewares(data)

        sql, args = self._build_insert(data)
        return execute(self._ds_id, sql, args)

    # noinspection PyMethodMayBeStatic
    def _process_data(self, data):
        # 判断 data 是否为dict，如果不是，则转换成dict
        if data is not None and not isinstance(data, dict):
            data = data.__dict__ if hasattr(data, "__dict__") else data.__dict__
        # 移除 value 为 None 的字段
        data = {k: v for k, v in data.items() if v is not None}
        return data

    def _build_insert(self, data: dict) -> tuple[str, tuple[Any, ...]]:
        placeholder = []
        keys = []
        values = []
        model_fields = [f.name for f in fields(self._model)]
        for k, v in data.items():
            if k in model_fields:
                keys.append(k)
                values.append(v)
                placeholder.append("?")

        if len(keys) == 0:
            raise ValueError("no valid field found")

        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'INSERT INTO {table}({",".join(keys)}) VALUES({",".join(placeholder)})'
        args = tuple(values)
        return sql, args

    def insert_bulk(self, data_list: list[dict]) -> int:
        if data_list is None or len(data_list) == 0:
            raise ValueError("data is required")

        sql, args = self._build_insert_bulk(data_list)
        return executemany(self._ds_id, sql, args)

    def _build_insert_bulk(self, data: list[dict]) -> tuple[str, list[tuple[any, ...]]]:
        placeholder = []
        keys = []
        model_fields = [f.name for f in fields(self._model)]
        for k, v in data[0].items():
            if k in model_fields:
                keys.append(k)
                placeholder.append("?")

        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'INSERT INTO {table}({",".join(keys)}) VALUES({",".join(placeholder)})'
        args = [tuple(datum[k] for k in keys) for datum in data]
        return sql, args

    def upsert(self, data: any, update_fields: list[str] | None = None) -> (int, int):
        if data is None:
            raise ValueError("null data")

        data = self._process_data(data)
        sql, args = self._build_upsert(data, update_fields)
        return execute(self._ds_id, sql, args)

    def _build_upsert(
        self, data: dict, update_fields: list[str] | None = None
    ) -> tuple[str, tuple[Any, ...]]:
        placeholder = []
        keys = []
        values = []
        model_fields = [f.name for f in fields(self._model)]

        for k, v in data.items():
            if k in model_fields:
                keys.append(k)
                values.append(v)
                placeholder.append("?")

        if len(keys) == 0:
            raise ValueError("no valid field found")

        # check if update_fields is a sub list of keys
        if update_fields is None:
            update_fields = keys
        else:
            if len(update_fields) == 0 or not all(
                field in keys for field in update_fields
            ):
                raise ValueError("update fields are not valid")

        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'INSERT INTO `{table}`({",".join(keys)}) VALUES({",".join(placeholder)})'
        if update_fields:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in update_fields])}'
        else:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in keys])}'

        args = tuple(values)
        return sql, args

    def upsert_bulk(
        self, data: list[dict], update_fields: list[str] | None = None
    ) -> int:
        if data is None or len(data) == 0:
            raise ValueError("data is required")

        sql, args = self._build_upsert_bulk(data, update_fields)
        return executemany(self._ds_id, sql, args)

    def _build_upsert_bulk(
        self, data: list[dict], update_fields: list[str] | None = None
    ) -> tuple[str, list[tuple[any, ...]]]:
        placeholder = []
        keys = []
        model_fields = [f.name for f in fields(self._model)]

        for k in data[0].keys():
            if k in model_fields:
                keys.append(k)
                placeholder.append("%s")

        if len(keys) == 0:
            raise ValueError("no valid field found")

        # check if update_fields is a sub list of keys
        if update_fields is None:
            update_fields = keys
        else:
            if len(update_fields) == 0 or not all(
                field in keys for field in update_fields
            ):
                raise ValueError("update fields are not valid")

        table = (
            self._table if self._database is None else f"{self._database}.{self._table}"
        )
        sql = f'INSERT INTO `{table}`({",".join(keys)}) VALUES({",".join(placeholder)})'
        if update_fields:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in update_fields])}'
        else:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in keys])}'

        args = [tuple(datum[k] for k in keys) for datum in data]
        return sql, args

    def _apply_before_insert_middlewares(self, data: dict):
        middlewares = get_middlewares(Middleware.BEFORE_INSERT)
        if middlewares is None or len(middlewares) == 0:
            return

        for middleware in middlewares:
            middleware(self, data)


def insert(
    table_or_cls: str | Type[T],
    database: str | None = None,
    ds_id: str = "default",
) -> Insert:
    if isinstance(table_or_cls, str):
        table = table_or_cls
    else:
        table = table_or_cls.__table_name__
    return Insert(table, database, ds_id)



from ._mysql_datasource import MysqlDataSource
from .protocols import IDataSource

data_sources: list[IDataSource] = []


def default_datasource() -> IDataSource | None:
    for ds in data_sources:
        if ds.get_id() == "default":
            return ds
    return None


def load_data_sources(config_dict: dict):
    data_source_config = config_dict.get("data_source")
    if data_source_config is None:
        raise ValueError("data_source config not found")

    for ds_id, ds_conf in data_source_config.items():
        if "dialect" not in ds_conf:
            raise ValueError("dialect is required")

        if "mysql" == ds_conf["dialect"]:
            mysql_ds = MysqlDataSource(
                ds_id=ds_id,
                host=ds_conf["host"],
                port=ds_conf["port"],
                user=ds_conf["user"],
                password=ds_conf["password"],
                database=ds_conf["database"],
            )
            data_sources.append(mysql_ds)
        else:
            raise ValueError(f'unsupported dialect: {ds_conf["dialect"]}')


def add_datasource(
    ds_id: str,
    dialect: str,
    host: str,
    port: int,
    user: str,
    password: str,
    database: str,
):
    if dialect is None or dialect == "":
        raise ValueError("dialect is required")
    if host is None or host == "":
        raise ValueError("host is required")
    if port is None or port <= 0:
        raise ValueError("port is required and must be greater than 0")
    if user is None or user == "":
        raise ValueError("user is required")
    if password is None or password == "":
        raise ValueError("password is required")
    if database is None or database == "":
        raise ValueError("database is required")

    for ds in data_sources:
        if ds.get_id() == ds_id:
            raise ValueError(f"data source with id {ds_id} already exists")

    if "mysql" == dialect:
        mysql_ds = MysqlDataSource(
            ds_id=ds_id,
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
        )
        data_sources.append(mysql_ds)
    else:
        raise ValueError(f"unsupported dialect: {dialect}")


def get_datasource(ds_id: str) -> IDataSource | None:
    for ds in data_sources:
        if ds.get_id() == ds_id:
            return ds
    return None



import uuid
from contextvars import ContextVar

from loguru import logger
from pymysql.connections import Connection

from ._datasources import default_datasource, get_datasource
from .protocols import IDataSource


class Tx:
    def __init__(self, ds_id: str | None = None, auto_commit: bool = True):
        self._auto_commit = auto_commit
        self._ds_id = ds_id
        self._datasource: IDataSource = (
            default_datasource() if ds_id is None else get_datasource(ds_id)
        )
        self._conn: Connection | None = None
        self._state = "new"
        self._tx_id = uuid.uuid4().hex
        logger.debug(
            f"Transaction [{self._tx_id}] created with auto_commit={auto_commit}"
        )

    def begin(self):
        if self._state == "new":
            if self._conn is not None:
                raise RuntimeError("Transaction already started")
            self._conn = self._datasource.create_connection()
            self._conn.begin()
            self._state = "begin"
        elif self._state == "begin" and not self._auto_commit:
            logger.debug(f"reuse existing transaction [{self._tx_id}]")
        elif self._state == "committed" or self._state == "rolled_back":
            raise RuntimeError("Transaction already completed")
        return self._conn

    def commit(self):
        if self._state != "begin":
            raise RuntimeError("Transaction not in a valid state to commit")
        if self._conn is None:
            raise RuntimeError("No connection to commit")
        self._conn.commit()
        logger.debug(f"Transaction [{self._tx_id}] committed")
        self._state = "committed"
        self._conn.close()

    def rollback(self):
        if self._state != "begin":
            raise RuntimeError("Transaction not in a valid state to rollback")
        if self._conn is None:
            raise RuntimeError("No connection to rollback")
        self._conn.rollback()
        logger.debug(f"Transaction [{self._tx_id}] rolled back")
        self._state = "rolled_back"
        self._conn.close()

    def is_auto_commit(self) -> bool:
        return self._auto_commit

    def is_valid(self) -> bool:
        return self._state == "new" or self._state == "begin"

    def ds_id(self) -> str:
        return self._ds_id


tx_context = ContextVar("sql_context", default=None)



from loguru import logger

from ._datasources import load_data_sources
from ._tx import tx_context, Tx


class Dorm:
    def __init__(self):
        self._init = False
        self._config_dict = None

    def is_initialized(self):
        return self._init

    def init(self, config_dict):
        self._config_dict = config_dict
        load_data_sources(config_dict)
        self._init = True
        logger.info("dorm initialized")

    # noinspection PyMethodMayBeStatic
    def begin(self, ds_id: str = "default"):
        tx = tx_context.get()
        if tx is not None and tx.is_valid():
            raise RuntimeError("Transaction already started")
        new_tx = Tx(ds_id=ds_id, auto_commit=False)
        tx_context.set(new_tx)

    # noinspection PyMethodMayBeStatic
    def commit(self):
        tx = tx_context.get()
        if tx is None:
            raise RuntimeError("No transaction to commit")

        if not tx.is_auto_commit():
            tx.commit()
            tx_context.set(None)
        else:
            logger.warning("Auto-commit is enabled, no explicit commit needed.")

    # noinspection PyMethodMayBeStatic
    def rollback(self):
        tx = tx_context.get()
        if tx is None:
            raise RuntimeError("No transaction to rollback")

        if not tx.is_auto_commit():
            tx.rollback()
            tx_context.set(None)
        else:
            logger.warning("Auto-commit is enabled, no explicit rollback needed.")


dorm = Dorm()



from ._datasources import add_datasource
from ._dorm import dorm
from ._entity import entity, vo
from ._initializer import init
from ._insert import insert, Insert
from ._middlewares import use_middleware
from ._models import models
from ._query import query, raw_query, Query, RawQuery
from ._raw import raw_select
from ._update import update, Update

__author__ = "melon"
__version__ = "0.7.2"

__all__ = [
    "init",
    "dorm",
    "query",
    "update",
    "insert",
    "raw_query",
    "entity",
    "vo",
    "Insert",
    "Query",
    "RawQuery",
    "Update",
    "use_middleware",
    "add_datasource",
    "models",
    "raw_select",
]
